// WBL $Revision: 1.6 $
//
// PNG drawing library for JavaScript.
// Copyright (C) 1999 by Roger E Critchlow Jr,
// Santa Fe, New Mexico, USA.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// The home page for Pnglets is http://www.elf.org/pnglets,
// a copy of the GNU GPL may be found at http://www.gnu.org/copyleft/gpl.html,
// Pnglets were inspired by and copied from gd1.3, http://www.boutell.com/gd,
// other parts were inspired by or copied from Tcl/Tk, http://www.scriptics.com,
// and some algorithms were taken from Foley & van Dam 2nd Edition.
//
// Thanks to Alex Vincent for pointing out the advantages of eliminating strict
// javascript warnings.
//

//WBL 17 Apr 2006 Make compatible with Mozilla Rhino 1.6R2
//W B Langdon 11 March 2004 Speed up.
//Use sparse coding 
//Modest recoding, and treating png[] and colours as numbers
//rather than strings yeilded further small speed ups
//
//on Mozilla 1.5 Mozilla/5.0 (Windows; U; Windows NT 5.0; en-US; rv:1.5) Gecko/20031007
//350mHz 12188mS for 701x300 png with approx 61 bits set. wbl_pnglet_test.html

// create a new Pnglet of specified width, height, and depth
// width and height are specified in pixels
// depth is really the number of palette entries
function Pnglet(width,height,depth) {
  this.width = width;   //WBL 17 Apr 2006|| 16;
  this.height = height; //WBL 17 Apr 2006|| 16;
  this.depth = Math.min(256, depth); //WBL 17 Apr 2006|| 16);

  // pixel data and row filter identifier size
  this.pix_size = height*(width+1);

  // deflate header, pix_size, block headers, adler32 checksum
  this.data_size = 2 + this.pix_size + 5*Math.floor((this.pix_size+0xffff-1)/0xffff) + 4;

  // offsets and sizes of Png chunks
  this.ihdr_offs = 0;					// IHDR offset and size
  this.ihdr_size = 4+4+13+4;
  this.plte_offs = this.ihdr_offs+this.ihdr_size;	// PLTE offset and size
  this.plte_size = 4+4+3*depth+4;
  this.trns_offs = this.plte_offs+this.plte_size;	// tRNS offset and size
  this.trns_size = 4+4+depth+4;
  this.idat_offs = this.trns_offs+this.trns_size;	// IDAT offset and size
  this.idat_size = 4+4+this.data_size+4;
  this.iend_offs = this.idat_offs+this.idat_size;	// IEND offset and size
  this.iend_size = 4+4+4;
  this.png_size = this.iend_offs+this.iend_size;	// total PNG size

  // array of numbers 0..255
  this.png = new Array(); //sparse! numbers not strings this.png_size);

  // functions for initializing data
  function initialize(png, offs, str) {
    for (var i = 1; i < arguments.length; i += 1)
      if (typeof arguments[i].length != "undefined")
        for (var j = 0; j < arguments[i].length; j += 1)
          png[offs++] = arguments[i].charCodeAt(j);
  };
  function byte2(w) { return String.fromCharCode((w>>8)&255, w&255); };
  function byte4(w) { return String.fromCharCode((w>>24)&255, (w>>16)&255, (w>>8)&255, w&255); };
  function byte2lsb(w) { return String.fromCharCode(w&255, (w>>8)&255); };

  //DONOT initialize everything to zero byte
  //for (var i = 0; i < this.png_size; i += 1)
  //  this.png[i] = String.fromCharCode(0);

  this.data = new Array(); //Pixels (sparse)

  // initialize non-zero elements
  initialize(this.png, this.ihdr_offs, byte4(this.ihdr_size-12), 'IHDR',
             byte4(width), byte4(height), String.fromCharCode(8, 3));
  initialize(this.png, this.plte_offs, byte4(this.plte_size-12), 'PLTE');
  initialize(this.png, this.trns_offs, byte4(this.trns_size-12), 'tRNS');
  initialize(this.png, this.idat_offs, byte4(this.idat_size-12), 'IDAT');
  initialize(this.png, this.iend_offs, byte4(this.iend_size-12), 'IEND');

  // initialize deflate header
  var header = ((8 + (7<<4)) << 8) | (3 << 6);
  header += 31 - (header % 31);
  initialize(this.png, this.idat_offs+8, byte2(header));

  // initialize deflate block headers
  for (i = 0; i*0xffff < this.pix_size; i += 1) {
	var size, bits;
	if (i + 0xffff < this.pix_size) {
      size = 0xffff;
      bits = String.fromCharCode(0);
	} else {
      size = this.pix_size - i*0xffff;
      bits = String.fromCharCode(1);
	}
	initialize(this.png, this.idat_offs+8+2+i*(5+0xffff), bits, byte2lsb(size), byte2lsb(~size));
  }

  // initialize palette hash
  this.palette = new Object();
  this.pindex = 0;
}

// version string/number
Pnglet.version = "19990427.0";

// test if coordinates are within bounds
Pnglet.prototype.inBounds = function(x,y) { return x >= 0 && x < this.width && y >= 0 && y < this.height; }

// clip an x value to the window width
Pnglet.prototype.clipX = function(x) { return (x < 0) ? 0 : (x >= this.width) ? this.width-1 : x ; }

// clip a y value to the window height
Pnglet.prototype.clipY = function(y) { return (y < 0) ? 0 : (y >= this.height) ? this.height-1 : y ; }


//needs to be faster! Overhead on function call & floor()
// compute the index into a png for a given pixel
Pnglet.prototype.index = function(x,y) {
  var i = y*(this.width+1)+x+1;
//  return this.idat_offs_15+i;
  return this.idat_offs+8+2+Math.floor((i/0xffff)+1)*5+i;
}

// make a color in a Pnglet
Pnglet.prototype.color = function(red, green, blue, alpha) {
  alpha = alpha >= 0 ? alpha : 255;
  var rgba = (((((alpha<<8)+red)<<8)+green)<<8)+blue;
  if ( typeof this.palette[rgba] == "undefined") {
	if (this.pindex == this.depth) return 0;
	this.palette[rgba] = this.pindex;
	this.png[this.plte_offs+8+this.pindex*3+0] = red;
	this.png[this.plte_offs+8+this.pindex*3+1] = green;
	this.png[this.plte_offs+8+this.pindex*3+2] = blue;
	this.png[this.trns_offs+8+this.pindex] = alpha;
	this.pindex += 1;
  }
  return this.palette[rgba];
}

// return true if this is a color
Pnglet.prototype.isColor = function(color) {
  return (color>=0 && color < this.depth);
/*
typeof(color) == 'string' &&
	color.length == 1 &&
	color.charCodeAt(0) >= 0 &&
	color.charCodeAt(0) < this.depth;
*/
}

// find the red, green, blue, or alpha value of a Pnglet color
Pnglet.prototype.red = function(color) { return this.png[this.plte_offs+8+color*3+0]; }
Pnglet.prototype.green = function(color) { return this.png[this.plte_offs+8+color*3+1]; }
Pnglet.prototype.blue = function(color) { return this.png[this.plte_offs+8+color*3+2]; }
Pnglet.prototype.alpha = function(color) { return this.png[this.trns_offs+8+color]; }

// draw a point or points
Pnglet.prototype.point = function(pointColor, x0, y0) {
  var a = arguments;
  this.pointNXY(pointColor, (a.length-1)/2, function(i) { return a[2*i+1]; }, function(i) { return a[2*i+2]; });
}

//sparse
Pnglet.prototype.setpixel = function(pointColor, x, y) {
  if(!this.data[y]) this.data[y] = new Array();
  this.data[y][x] = pointColor;
  this.png[this.index(x,y)] = pointColor; //To avoid slow index() store data twice
}
Pnglet.prototype.pointNXY = function(pointColor, n, x, y) {
  if ( ! this.isColor(pointColor))
	return;
  for (var i = 0; i < n; i += 1) {
	var x1 = x(i), y1 = y(i);
	if (this.inBounds(x1,y1)) this.setpixel(pointColor,x1,y1);
  }
}

// read a pixel 
Pnglet.prototype.getPoint = function(x,y) { return this.inBounds(x,y) ? this.data[y][x] : 0; }

// draw a horizontal line
Pnglet.prototype.horizontalLine = function(lineColor, x1, x2, y) {
  if ( ! this.isColor(lineColor))
	return;
  x1 = this.clipX(x1);
  x2 = this.clipX(x2);
  var x;
  if (x1 < x2)
	for (x = x1; x <= x2; x += 1) this.setpixel(lineColor,x,y);
  else
	for (x = x2; x <= x1; x += 1) this.setpixel(lineColor,x,y);
}

// draw a vertical line
Pnglet.prototype.verticalLine = function(lineColor, x, y1, y2) {
  if ( ! this.isColor(lineColor))
	return;
  y1 = this.clipY(y1);
  y2 = this.clipY(y2);
  var y;
  if (y1 < y2)
	for (y = y1; y <= y2; y += 1) this.setpixel(lineColor,x,y);
  else
	for (y = y2; y <= y1; y += 1) this.setpixel(lineColor,x,y);
}

// draw a general line
Pnglet.prototype.generalLine = function(lineColor, x1, y1, x2, y2) {
  if ( ! this.isColor(lineColor))
	return;
  var dx = Math.abs(x2-x1), dy = Math.abs(y2-y1);
  var incr1, incr2, d, x, y, xend, yend, xdirflag, ydirflag, xinc, yinc;
  if (dy <= dx) {
	d = 2*dy - dx;
	incr1 = 2*dy;
	incr2 = 2 * (dy - dx);
	if (x1 > x2) {
      x = x2;
      y = y2;
      ydirflag = -1;
      xend = x1;
	} else {
      x = x1;
      y = y1;
      ydirflag = 1;
      xend = x2;
	}
	yinc = (((y2 - y1) * ydirflag) > 0) ? 1 : -1;
	this.point(lineColor, x, y);
	while (x++ < xend) {
      if (d < 0) {
		d += incr1;
      } else {
		y += yinc;
		d += incr2;
      }
      this.point(lineColor, x, y);
	}
  } else {			/* dy > dx */
	d = 2*dx - dy;
	incr1 = 2*dx;
	incr2 = 2 * (dx - dy);
	if (y1 > y2) {
      y = y2;
      x = x2;
      yend = y1;
      xdirflag = -1;
	} else {
      y = y1;
      x = x1;
      yend = y2;
      xdirflag = 1;
	}
	xinc = (((x2 - x1) * xdirflag) > 0) ? 1 : -1;
	this.point(lineColor, x, y);
	while (y++ < yend) {
      if (d < 0) {
		d += incr1;
      } else {
		x += xinc;
		d += incr2;
      }
      this.point(lineColor, x, y);
	}
  }
}

// draw a line
Pnglet.prototype.line = function(lineColor, x0, y0) {
  var a = arguments;
  this.lineNXY(lineColor, (a.length-1)/2, function(i) { return a[2*i+1]; }, function(i) { return a[2*i+2]; });
}

Pnglet.prototype.lineNXY = function(lineColor, n, x, y) {
  if ( ! this.isColor(lineColor))
	return;
  var x1 = x(0), y1 = y(0);
  for (var i = 1; i < n; i += 1) {
	var x2 = x(i), y2 = y(i);
	if (x1 == x2)
      this.verticalLine(lineColor, x1, y1, y2);
	else if (y1 == y2)
      this.horizontalLine(lineColor, x1, x2, y1);
	else
      this.generalLine(lineColor, x1, y1, x2, y2);
	x1 = x2;
	y1 = y2;
  }
}

// draw a polygon 
Pnglet.prototype.polygonA = function(outlineColor, fillColor, X, Y) {
  this.polygonNXY(outlineColor, fillColor, X.length, function(i) {return X[i];}, function(i) {return Y[i];});
}

// draw a polygon
Pnglet.prototype.polygon = function(outlineColor, fillColor, x1, y1) {
  var a = arguments;
  this.polygonNXY(outlineColor, fillColor, (a.length-2)/2, function(i) {return a[2*i+2];}, function(i) {return a[2*i+3];});
}

Pnglet.prototype.polygonNXY = function(outlineColor, fillColor, n, x, y) {
  if (n <= 0)
	return;
  if (this.isColor(fillColor))
	this.concaveNXY(fillColor, n, x, y);
  if (this.isColor(outlineColor))
	this.lineNXY(outlineColor, n+1, function(i) { return x(i%n); }, function(i) { return y(i%n); });
}

/*
 * Concave Polygon Scan Conversion
 * by Paul Heckbert
 * from "Graphics Gems", Academic Press, 1990
 */
Pnglet.prototype.concaveNXY = function(fillColor, n, ptx, pty) {
  function Edge(ex, edx, ei) {	/* a polygon edge */
	this.x = ex;	/* x coordinate of edge's intersection with current scanline */
	this.dx = edx;	/* change in x with respect to y */
	this.i = ei;	/* edge number: edge i goes from pt[i] to pt[i+1] */
  };
  function cdelete(di) {	/* remove edge i from active list */
    for (var j = 0; j < active.length; j += 1)
      if (active[j].i == di)
        active.splice(j, 1);
  };
  function cinsert(ii, iy) {	/* append edge i to end of active list */
    var ij = ii<n-1 ? ii+1 : 0;
    var px, py, qx, qy;
    if (pty(ii) < pty(ij)) {
      px = ptx(ii); py = pty(ii);
      qx = ptx(ij); qy = pty(ij);
    } else {
      px = ptx(ij); py = pty(ij);
      qx = ptx(ii); qy = pty(ii);
    }
    /* initialize x position at intersection of edge with scanline y */
    var dx = (qx-px)/(qy-py);
    active.push(new Edge(dx*(iy+.5-py)+px, dx, ii));
  };

  var ind = new Array(n);		/* list of vertex indices, sorted by pt[ind[j]].y */
  var active = new Array(0);		/* start with an empty active list */

  /* create y-sorted array of indices ind[k] into vertex list */
  for (var k = 0; k < n; k += 1) ind[k] = k;
  ind.sort(function(i1, i2) { return pty(i1) <= pty(i2) ? -1 : 1; });
  k = 0;                        /* ind[k] is next vertex to process */
  var y0 = Math.max(0, Math.ceil(pty(ind[0])+.5));			/* ymin of polygon */
  var y1 = Math.min(this.height, Math.floor(pty(ind[n-1])-.5));	/* ymax of polygon */

  for (var y = y0; y <= y1; y += 1) {		/* step through scanlines */
	/* scanline y is at y+.5 in continuous coordinates */

	/* check vertices between previous scanline and current one, if any */
	for (; k<n && pty(ind[k]) <= y+.5; k += 1) {
      /* to simplify, if pt.y=y+.5, pretend it's above */
      /* invariant: y-.5 < pt[i].y <= y+.5 */
      var i = ind[k];	

      /*
       * insert or delete edges before and after vertex i (i-1 to i,
       * and i to i+1) from active list if they cross scanline y
       */
      var j = (i-1+n)%n;		/* vertex previous to i */
      if (pty(j) <= y-.5)	{	/* old edge, remove from active list */
		cdelete(j);
      } else if (pty(j) > y+.5) {	/* new edge, add to active list */
		cinsert(j, y);
      }
      if (i != ind[k]) {
		alert("Your browser's implementation of JavaScript is seriously broken,\n"+
		      "as in variables are changing value of their own volition.\n"+
		      "You should upgrade to a newer version browser.");
		return;
      }
      j = (i+1)%n;		/* vertex next after i */
      if (pty(j) <= y-.5)	{	/* old edge, remove from active list */
		cdelete(i);
      } else if (pty(j) > y+.5) {	/* new edge, add to active list */
		cinsert(i, y);
      }
	}

	/* sort active edge list by active[j].x */
	active.sort(function(u,v) { return u.x <= v.x ? -1 : 1; });

	/* draw horizontal segments for scanline y */
	for (j = 0; j < active.length; j += 2) {	/* draw horizontal segments */
      /* span 'tween j & j+1 is inside, span tween j+1 & j+2 is outside */
      var xl = Math.ceil(active[j].x+.5);		/* left end of span */
      if (xl<0) xl = 0;
      var xr = Math.floor(active[j+1].x-.5);	/* right end of span */
      if (xr>this.width-1) xr = this.width-1;
      if (xl<=xr)
		this.horizontalLine(fillColor, xl, xr, y);	/* draw pixels in span */
      active[j].x += active[j].dx;	/* increment edge coords */
      active[j+1].x += active[j+1].dx;
	}
  }
}

// draw a rectangle
Pnglet.prototype.rectangle = function(outlineColor, fillColor, x0,y0,x1,y1) {
  if (this.isColor(fillColor))
	for (var y = y0; y < y1; y += 1)
      this.horizontalLine(fillColor, x0+1, x1-2, y);
  if (this.isColor(outlineColor)) {
	this.horizontalLine(outlineColor, x0, x1-1, y0);
	this.horizontalLine(outlineColor, x0, x1-1, y1-1);
	this.verticalLine(outlineColor, x0, y0, y1-1);
	this.verticalLine(outlineColor, x1-1, y0, y1-1);
  }
}
	
// draw an arc
Pnglet.prototype.arc = function(outlineColor, cx,cy,w,h, s,e) {
  var p = this.midpointEllipse(cx,cy, w,h, s,e);
  function x(i) { return p[i*2]; };
  function y(i) { return p[i*2+1]; };
  this.lineNXY(outlineColor, p.length/2, x, y);
}

// draw an oval
Pnglet.prototype.oval = function(outlineColor, fillColor, cx,cy,w,h) {
  var p = this.midpointEllipse(cx,cy, w,h, 0,359);
  function x(i) { return p[i*2]; };
  function y(i) { return p[i*2+1]; };
  this.polygonNXY(outlineColor, fillColor, p.length/2, x, y);
}

// draw an arc with chord
Pnglet.prototype.chord = function(outlineColor, fillColor, cx,cy,w,h, s,e) {
  var p = this.midpointEllipse(cx,cy, w,h, s,e);
  function x(i) { return p[i*2]; };
  function y(i) { return p[i*2+1]; };
  this.polygonNXY(outlineColor, fillColor, p.length/2, x, y);
}

// draw an arc with pieslice
Pnglet.prototype.pieslice = function(outlineColor, fillColor, cx,cy,w,h, s,e) {
  var p = this.midpointEllipse(cx,cy, w,h, s,e);
  p[p.length] = cx;
  p[p.length] = cy;
  function x(i) { return p[i*2]; };
  function y(i) { return p[i*2+1]; };
  this.polygonNXY(outlineColor, fillColor, p.length/2, x, y);
}

// oval arcs
// generate points of oval circumference
// midpoint ellipse, Foley & van Dam, 2nd Edition, p. 90, 1990
Pnglet.prototype.midpointEllipse = function(cx,cy, w,h, s,e) {
  var a = Math.floor(w/2), b = Math.floor(h/2);
  var a2 = a*a, b2 = b*b, x = 0, y = b;
  var d1 = b2 - a2*b + a2/4;
  cx = Math.floor(cx);
  cy = Math.floor(cy);
  var p = new Array();

  // quadrant I, anticlockwise
  p.push(x,-y);
  while (a2*(y-1/2) > b2*(x+1)) {
	if (d1 < 0) {
      d1 += b2*(2*x+3);
	} else {
      d1 += b2*(2*x+3) + a2*(-2*y + 2);
      y -= 1;
	}
	x += 1;
	p.unshift(x,-y);
  }
  var d2 = b2*(x+1/2)*(x+1/2) + a2*(y-1)*(y-1) - a2*b2;
  while (y > 0) {
	if (d2 < 0) {
      d2 += b2*(2*x+2) + a2*(-2*y+3);
      x += 1;
	} else {
      d2 += a2*(-2*y+3);
	}
	y -= 1;
	p.unshift(x,-y);
  }
  // quadrant II, anticlockwise
  var n4 = p.length;
  for (var i = n4-4; i >= 0; i -= 2)
    p.push(-p[i], p[i+1]);
  // quadrants III and IV, anticlockwise
  var n2 = p.length;
  for (i = n2-4; i > 0; i -= 2)
    p.push(p[i], -p[i+1]);

  // compute start and end indexes from start and extent
  e %= 360;
  if (e < 0) {
	s += e;
	e = -e;
  }
  s %= 360;
  if (s < 0)
	s += 360;
  var is = Math.floor(s/359 * p.length/2);
  var ie = Math.floor(e/359 * p.length/2)+1;
  p = p.slice(is*2).concat(p.slice(0, is*2)).slice(0, ie*2);

  // displace to center
  for (i = 0; i < p.length; i += 2) {
	p[i] += cx;
	p[i+1] += cy;
  }
  return p;
}

// fill a region 
// from gd1.3 with modifications
Pnglet.prototype.fill = function(outlineColor,fillColor,x,y) {
  if (outlineColor) {			// fill to outline color
	/* Seek left */
	var leftLimit = -1;
	for (var i = x; i >= 0 && this.getPoint(i, y) != outlineColor; i -= 1)
      leftLimit = i;
	
	if (leftLimit == -1)
      return;

	/* Seek right */
	var rightLimit = x;
	for (i = (x+1); i < this.width && this.getPoint(i, y) != outlineColor; i += 1)
      rightLimit = i;

	/* fill extent found */
	this.horizontalLine(fillColor, leftLimit, rightLimit, y);

	/* Seek above and below */
	for (var dy = -1; dy <= 1; dy += 2) {
      if (this.inBounds(x,y+dy)) {
		var lastBorder = 1;
		for (i = leftLimit; i <= rightLimit; i++) {
          var c = this.getPoint(i, y+dy);
          if (lastBorder) {
			if ((c != outlineColor) && (c != fillColor)) {	
              this.fill(outlineColor, fillColor, i, y+dy);
              lastBorder = 0;
			}
          } else if ((c == outlineColor) || (c == fillColor)) {
			lastBorder = 1;
          }
		}
      }
	}
    
  } else {			// flood fill color at x, y 
	/* Test for completion */
	var oldColor = this.getPoint(x, y);
	if (oldColor == fillColor)
      return;

	/* Seek left */
	leftLimit = (-1);
	for (i = x; i >= 0 && this.getPoint(i, y) == oldColor; i--)
      leftLimit = i;

	if (leftLimit == -1)
      return;

	/* Seek right */
	rightLimit = x;
	for (i = (x+1); i < this.width && this.getPoint(i, y) == oldColor; i++)
      rightLimit = i;

	/* Fill extent found */
	this.horizontalLine(fillColor, leftLimit, rightLimit, y);

	/* Seek above and below */
	for (dy = -1; dy <= 1; dy += 2) {
      if (this.inBounds(x,y+dy)) {
		lastBorder = 1;
		for (i = leftLimit; i <= rightLimit; i++) {
          c = this.getPoint(i, y+dy);
          if (lastBorder) {
			if (c == oldColor) {	
              this.fill(null, fillColor, i, y+dy);
              lastBorder = 0;
			}
          } else if (c != oldColor) {
			lastBorder = 1;
          }
		}
      }
	}
  }
}

// smoothed points
Pnglet.prototype.smoothPoint = function(smoothSteps, pointColor, x0, y0) {
  var a = arguments, self = this, n = (a.length-2)/2;
  this.smooth(smoothSteps,
              function(n, x, y) { self.pointNXY(pointColor, n, x, y); },
              n,
              function(i) { return a[2*i+2]; },
              function(i) { return a[2*i+3]; });
}

// smoothed polyline
Pnglet.prototype.smoothLine = function(smoothSteps, lineColor, x0, y0) {
  var a = arguments, self = this, n = (a.length-2)/2;
  this.smooth(smoothSteps,
              function(n, x, y) { self.lineNXY(lineColor, n, x, y); },
              n,
              function(i) { return a[2*i+2]; },
              function(i) { return a[2*i+3]; });
}

// smoothed polygon
Pnglet.prototype.smoothPolygon = function(smoothSteps, outlineColor, fillColor, x0, y0) {
  var a = arguments, self = this, n = (a.length-3)/2 + 1;
  this.smooth(smoothSteps,
              function(n, x, y) { self.polygonNXY(outlineColor, fillColor, n, x, y); },
              n,
              function(i) { return a[2*(i%(n-1))+3]; },
              function(i) { return a[2*(i%(n-1))+4]; });
}
    
// generate smoothSteps points for the line segment connecting
// each consecutive pair of points in x(i), y(i).
// adapted from the source for tk8.1b3, http://www.scriptics.com
Pnglet.prototype.smooth = function(smoothSteps, fNXY, n, x, y) {
  var control = new Array(8);
  var outputPoints = 0;
  var dblPoints = new Array();

  // compute numSteps of smoothed points
  // according to the basis in control[]
  // placing points into coordPtr[coordOff]
  function smoothPoints(control, numSteps, coordPtr, coordOff) {
	for (var i = 1; i <= numSteps; i++, coordOff += 2) {
      var t = i/numSteps, t2 = t*t, t3 = t2*t,
		u = 1.0 - t, u2 = u*u, u3 = u2*u;
      coordPtr[coordOff+0] = control[0]*u3 + 3.0 * (control[2]*t*u2 + control[4]*t2*u) + control[6]*t3;
      coordPtr[coordOff+1] = control[1]*u3 + 3.0 * (control[3]*t*u2 + control[5]*t2*u) + control[7]*t3;
	}
  };

  /*
   * If the curve is a closed one then generate a special spline
   * that spans the last points and the first ones.  Otherwise
   * just put the first point into the output.
   */

  var closed = (x(0) == x(n-1)) && (y(0) == y(n-1));
  if (closed) {
	control[0] = 0.5*x(n-2) + 0.5*x(0);
	control[1] = 0.5*y(n-2) + 0.5*y(0);
	control[2] = 0.167*x(n-2) + 0.833*x(0);
	control[3] = 0.167*y(n-2) + 0.833*y(0);
	control[4] = 0.833*x(0) + 0.167*x(1);
	control[5] = 0.833*y(0) + 0.167*y(1);
	control[6] = 0.5*x(0) + 0.5*x(1);
	control[7] = 0.5*y(0) + 0.5*y(1);
	dblPoints[2*outputPoints+0] = control[0];
	dblPoints[2*outputPoints+1] = control[1];
	outputPoints += 1;
	smoothPoints(control, smoothSteps, dblPoints, 2*outputPoints);
	outputPoints += smoothSteps;
  } else {
	dblPoints[2*outputPoints+0] = x(0);
	dblPoints[2*outputPoints+1] = y(0);
	outputPoints += 1;
  }

  for (var i = 2; i < n; i += 1) {
	var j = i - 2;
	/*
	 * Set up the first two control points.  This is done
	 * differently for the first spline of an open curve
	 * than for other cases.
	 */
	if ((i == 2) && !closed) {
      control[0] = x(j);
      control[1] = y(j);
      control[2] = 0.333*x(j) + 0.667*x(j+1);
      control[3] = 0.333*y(j) + 0.667*y(j+1);
	} else {
      control[0] = 0.5*x(j) + 0.5*x(j+1);
      control[1] = 0.5*y(j) + 0.5*y(j+1);
      control[2] = 0.167*x(j) + 0.833*x(j+1);
      control[3] = 0.167*y(j) + 0.833*y(j+1);
	}

	/*
	 * Set up the last two control points.  This is done
	 * differently for the last spline of an open curve
	 * than for other cases.
	 */

	if ((i == (n-1)) && !closed) {
      control[4] = .667*x(j+1) + .333*x(j+2);
      control[5] = .667*y(j+1) + .333*y(j+2);
      control[6] = x(j+2);
      control[7] = y(j+2);
	} else {
      control[4] = .833*x(j+1) + .167*x(j+2);
      control[5] = .833*y(j+1) + .167*y(j+2);
      control[6] = 0.5*x(j+1) + 0.5*x(j+2);
      control[7] = 0.5*y(j+1) + 0.5*y(j+2);
	}

	/*
	 * If the first two points coincide, or if the last
	 * two points coincide, then generate a single
	 * straight-line segment by outputting the last control
	 * point.
	 */

	if (((x(j) == x(j+1)) && (y(j) == y(j+1)))
        || ((x(j+1) == x(j+2)) && (y(j+1) == y(j+2)))) {
	  dblPoints[2*outputPoints+0] = control[6];
	  dblPoints[2*outputPoints+1] = control[7];
	  outputPoints += 1;
	  continue;
	}

	/*
	 * Generate a Bezier spline using the control points.
	 */
	smoothPoints(control, smoothSteps, dblPoints, 2*outputPoints);
	outputPoints += smoothSteps;
  }

  // draw the points
  // anonymous functions don't work here
  // they result in "undefined" point values
  function myx(i) { return Math.round(dblPoints[2*i]); }
    function myy(i) { return Math.round(dblPoints[2*i+1]); }
    fNXY(outputPoints, myx, myy);
}

// output a PNG string
Pnglet.prototype.output = function(nset) {
  // output translations
  function initialize(png, offs, str) {
	for (var i = 1; i < arguments.length; i += 1)
      if (typeof arguments[i].length != "undefined")
        for (var j = 0; j < arguments[i].length; j += 1)
          png[offs++] = arguments[i].charCodeAt(j);
  }
    function byte4(w) { return String.fromCharCode((w>>24)&255, (w>>16)&255, (w>>8)&255, w&255); }

var start = new Date().getTime();
//defaultStatus = "adler";

    // compute adler32 of output pixels + row filter bytes
//    var BASE = 65521; /* largest prime smaller than 65536 */
//    var NMAX = 5552;  /* NMAX is the largest n such that 255n(n+1)/2 + (n+1)(BASE-1) <= 2^32-1 */
/*********************************************************************
  var s1 = 1;
  var s2 = 0;
  var n = NMAX;
  for (var y = 0; y < this.height; y += 1)
	for (var x = -1; x < this.width; x += 1) {
    if(this.png[this.index(x,y)]) {
      s1 += this.png[this.index(x,y)].charCodeAt(0);
    }//endif png[] exists
      s2 += s1;
      if ((n -= 1) == 0) {
		s1 %= BASE;
		s2 %= BASE;
		n = NMAX;
      }
	}
  s1 %= BASE;
  s2 %= BASE;
*********************************************************************/
  {
//WBL convert to JavaScript and sparse encoding
//http://heanet.dl.sourceforge.net/sourceforge/mhash/mhash-0.8.18.tar.gz
//cf void mhash_adler32(word32 * adler, const void *given_buf, int len)
//Based on Copyright (C) 2001 Nikos Mavroyanopoulos
  var s1 = 1;
  var s2 = 0;
  var w = this.width;
  var h = this.height
  for (var y = 0;  y < h; y++) {
    s2 += s1;	        //deal with x =-1 as special case
    if(s2 >= 65521)	/* same warning applies here, too */
      s2 -= 65521;
  for (var x = 0; x < w; x++) {
    var q=this.data[y];
    var p;
    if(q && (p=q[x])) {
      if(!nset[p]) nset[p]=1; else nset[p]++;
      s1 += p;
      if(s1 >= 65521)	/* using modulo took about 7 times longer on Mavroyanopoulos CPU! */
        s1 -= 65521;	/* WARNING: it's meant to be >= 65521, not just > ! */
    }//endif png[] exists
    s2 += s1;
    if(s2 >= 65521)	/* same warning applies here, too */
      s2 -= 65521;
    }//endfor x
    }//endfor y
    /*if(s1!=s1_ || s2!=s2_) alert("3adler32 missmatch! "+s1+"=="+s1+" "+s2+"=="+s2+"?");*/
  }
  initialize(this.png, this.idat_offs+this.idat_size-8, byte4((s2 << 16) | s1));

//defaultStatus += "32 "+((new Date().getTime())-start);

  // compute crc32 of the PNG chunks
  function crc32(png, offs, size) {
	var crc = -1;		// initialize crc
        var end = offs+size-4;
	for (var i = offs+4; i < end; i++) {
	  if(!png[i])
	    crc = crc32_table[crc & 0xff] ^ (crc >>> 8);
	  else
	    crc = crc32_table[(crc ^ png[i]) & 0xff] ^ (crc >>> 8);
        }
	initialize(png, offs+size-4, byte4(crc ^ -1));
  }

    crc32(this.png, this.ihdr_offs, this.ihdr_size);
//defaultStatus += " ihdr"+((new Date().getTime())-start);
  crc32(this.png, this.plte_offs, this.plte_size);
//defaultStatus += " plte"+((new Date().getTime())-start);
  crc32(this.png, this.trns_offs, this.trns_size);
//defaultStatus += " trns"+((new Date().getTime())-start);
  crc32(this.png, this.idat_offs, this.idat_size);
//defaultStatus += " idat"+((new Date().getTime())-start);
  crc32(this.png, this.iend_offs, this.iend_size);
//defaultStatus += " iend"+((new Date().getTime())-start);

  // convert PNG to string
  //return "\211PNG\r\n\032\n"+this.png.join('');

  var z = String.fromCharCode(0);
  var t ="";
  for (var i = 0; i <this.png_size;i++){t += (!this.png[i])? z : String.fromCharCode(this.png[i]);}
//defaultStatus += " uspr_1_"+((new Date().getTime())-start);

  // convert PNG to string
  return "\211PNG\r\n\032\n"+t;
}

/* Table of CRCs of all 8-bit messages. */
crc32_table = new Array(256);
for (var n = 0; n < 256; n++) {
  var c = n;
  for (var k = 0; k < 8; k++) {
    if (c & 1)
      c = -306674912 ^ ((c >> 1) & 0x7fffffff);
    else
      c = (c >> 1) & 0x7fffffff;
  }
  crc32_table[n] = c;
}
// D01 Lindenmayer System $Revision: 1.26 $
// W. B. Langdon
// http://www.cs.ucl.ac.uk/staff/W.Langdon
// Written on Jan 2004 (based on lsys.perl r1.11)
// itself copied from http://hjem.get2net.dk/bnielsen/lsystems.html 15 Dec 2001
// You may use this if you keep this copyright notice intact
// --------------------
// Tested with Mozilla Firebird and IE5.00. 
// Pnglet does NOT work with IE5.0 or IE6.0
// Netscape 4 not tried with Pnglet. NS4 unstable with pfeiffer_local.
//   (Also NS4 gives extra vertical lines at lefthand margin of XBM images).

//WBL 17 Apr 2006   Ensure compatible with Rhino R1.6r2
//WBL 11 Nov 2005   Ensure do not generate bad SVG when c2 is not valid
//WBL 25 Aug 2005   (Nearly) full LD0 support. 
//                  Ensure SVG polygon sets bounding box
//WBL 31 July 2005  Start adding SVG support
//WBL 24 Mar 2004   Merge PNG and XBM versions
//WBL 13 March 2004 Support colour and more Turtle commands
//                  Trap popping either empty stack

// move nset from pixmap to new xbm_pixmap argument dims
// create separate rotate_pixmap
// Allow evalstring to take bounding box inputs 7
// Allow planar figure to be out of X-Y plane
// IE5 no "const" correct used of undefined

//hashtables see http://www.crockford.com/javascript/survey.html
//ppm see http://en2.wikipedia.org/wiki/Portable_Pixmap_File_Format
//        http://www-info2.informatik.uni-wuerzburg.de/mitarbeiter/wolfram/lehre/bildformate.html

//I have not been able to find any justification for need to invert
//bit map so trees grow upwards but it appears to be needed.


        /*http://www.adobe.com/svg/source/transform_js.html
	dont quite see how to do this yet.
        * This function sets the transform attribute of the "gradient_star"
        * SVG element to the indicated value.
        *
        function set_transform (value)
        {
                var svgobj;
                var svgdoc = document.transform.getSVGDocument();

                svgobj = svgdoc.getElementById('gradient_star');
                svgobj.setAttribute ('transform', value);
                if (document.transf.transform.value != value)
                        document.transf.transform.value = value;
        }
        * This function appends the indicated value to the existing
        * transform attribute of the "gradient_star" SVG element.
        *
        function add_transform (value)
        {
                var svgobj;
                var svgdoc = document.transform.getSVGDocument();

                svgobj = svgdoc.getElementById('gradient_star');
                var cur_transform = svgobj.getAttribute ('transform');
                svgobj.setAttribute ('transform', cur_transform + ' ' + value);
                document.transf.transform.value = cur_transform + ' ' + value;
        }
*/
//could not make sense of SVG skewX, skewY transforms
//so do simple thing.
function rotate_svg(pixmap,angle) {
  newpix = new Array; //y x for efficiency
  newpix.minx=0;
  newpix.maxx=0;
  newpix.miny=0;
  newpix.maxy=0;

    var M = glRotate(angle,0,0,1); //matrix for rotation about z-axis
    //following will tend to overestimate bounds
    set_bounds(M[0][0]*pixmap.minx+M[0][1]*pixmap.miny,
	       M[1][0]*pixmap.minx+M[1][1]*pixmap.miny,newpix);
    set_bounds(M[0][0]*pixmap.minx+M[0][1]*pixmap.maxy,
	       M[1][0]*pixmap.minx+M[1][1]*pixmap.maxy,newpix);
    set_bounds(M[0][0]*pixmap.maxx+M[0][1]*pixmap.miny,
	       M[1][0]*pixmap.maxx+M[1][1]*pixmap.miny,newpix);
    set_bounds(M[0][0]*pixmap.maxx+M[0][1]*pixmap.maxy,
	       M[1][0]*pixmap.maxx+M[1][1]*pixmap.maxy,newpix);


    //still in cartesian coordinates
    var bbox= "<!-- rotated by "+angle+" degrees "+
	      " "+Date()+"\n"+
              M[0][0]+","+M[0][1]+","+M[1][0]+","+M[1][1]+"\n"+
	      "new bbox "+
	      newpix.minx+" "+newpix.miny+" "+newpix.maxx+" "+newpix.maxy+"-->"

    var trans = "transform=\""+
                "translate("+(-newpix.minx)+","+newpix.maxy+") "+
	        "scale(1,-1) "+
	        "rotate("+angle+")\">";

    var h=pixmap.txt.indexOf("<g id=\"top\"");
    var h2=pixmap.txt.indexOf(">",h);
    if(h<0||h2<=h) {
      alert("no <g id=\"top\" in SVG,"+h+" "+h2); 
      return pixmap;}
    var x =pixmap.txt.indexOf("transform=",h);
    var header=pixmap.txt.substring(0,x);
    var x2=pixmap.txt.indexOf("\">",x);
    var tail = pixmap.txt.substring(h2+2);
    newpix.txt=header+trans+"\n"+bbox+"\n"+tail;
    return newpix;
}

function rotate(pixmap,M) {/*bitmap_output*/
  newpix = new Array; //y x for efficiency
  newpix.minx=0;
  newpix.maxx=0;
  newpix.miny=0;
  newpix.maxy=0;
  //noneed to center done externally by moving DIV to retain centering

  if(pixmap.polygon) {
    newpix.polygon = new Array(pixmap.polygon.length);
    for(var n = 0; n<pixmap.polygon.length; n++) {
      var poly  = pixmap.polygon[n];
      var newp  = newpix.polygon[n] = new Object;
      newp.line = poly.line;
      newp.fill = poly.fill;
      newp.x    = new Array(poly.x.length);
      newp.y    = new Array(poly.y.length);
      for(var i=0; i<poly.x.length;i++) {
        newp.x[i] = Math.round(M[0][0]*poly.x[i] + M[0][1]*poly.y[i]);
	newp.y[i] = Math.round(M[1][0]*poly.x[i] + M[1][1]*poly.y[i]);
        set_bounds(newp.x[i],newp.y[i],newpix);
      }
    }
//  defaultStatus += "rot"+newpix.polygon[0].line+","+newpix.polygon[0].fill;
  }

  for(var y in pixmap) {
    var m01y = M[0][1]*y;
    var m11y = M[1][1]*y;
    for(var x in pixmap[y]) {
      var x_ = Math.round(M[0][0]*x + m01y);
      var y_ = Math.round(M[1][0]*x + m11y);
      set_bounds(x_,y_,newpix);
      if(newpix[y_]==null) {newpix[y_] = new Array;}
      newpix[y_][x_] = pixmap[y][x];
    }
  }  
  return newpix;
}

function pnglet_data(pixmap,dims,cmap) {
 var timeout_start = new Date().getTime();
 var width  = 1+pixmap.maxx-pixmap.minx;
 var height = 1+pixmap.maxy-pixmap.miny;
//defaultStatus += "png_pixmap("+width+"x"+height+",dims)";

 //Limit size of image so pnglet.output does not exceed timeout limit
 //(Assume 50,000 pixels per elapsed second)
 var max_h = Math.ceil(limit*50/width); 
 if(limit!=0 && max_h<height) {
   height = max_h;
   timedout |= 4;
// defaultStatus += "to:"+timedout+" "+max_h;
 }
 var colours = new Array((cmap)? cmap.length+2 : 2);
//defaultStatus += "#"+colours.length;
 var png = new Pnglet(width, height, colours.length);
 colours[0] = png.color(0,0,0,0);  //Actually creates the colour palette
 colours[1] = png.color(255,255,255);
 for(var k=0; k<colours.length-2;k++) {
   colours[k+2] = png.color(cmap[k][0],cmap[k][1],cmap[k][2]); }

 //bit of a hack since polygons cannot be in right order wrt pixels
 if(pixmap.polygon) {
   for(var n = 0; n<pixmap.polygon.length; n++) {
     if(timeout(timeout_start)) {
       timedout |= 8;
//     defaultStatus += "top:"+timedout+" "+n;
       break;
     }
     var poly = pixmap.polygon[n];
     var count= poly.x.length;
     if(count>2) { //ignore badly formed polygons
       var c1   = colours[poly.line];
       var c2   = colours[poly.fill];
       X        = new Array(count);
       Y        = new Array(count);
//     defaultStatus += "pg"+c1+","+c2+"(";
       for(var i=0; i<count;i++) {//diff XY-conventions
         X[i] = poly.x[i] - pixmap.minx
	 Y[i] = pixmap.maxy - poly.y[i];
//	 defaultStatus += ""+X[i]+","+Y[i]+" "
       }
//     defaultStatus += ")";
       png.polygonA(c1,c2,X,Y);
//     if(count!=3) alert("Calling polygon with "+X.length+" points");
//     png.polygonNXY(c1,c2,count,
//                    function(i) {return X[i];},
//                    function(i) {return Y[i];});
     }
   }
 }

 for(y = pixmap.maxy ; y > pixmap.maxy-height; y--) {//start at top
   if(timeout(timeout_start)) {
     timedout |= 8;
//   defaultStatus += "toy:"+timedout+" "+(y-pixmap.maxy+height);
     break;
   }
   if(pixmap[y]) {
     for(var x in pixmap[y]) { //diff XY-conventions
       png.point(colours[pixmap[y][x]], x-pixmap.minx, pixmap.maxy-y);
     }//endfor
   }//endif
 }

if((timedout&(8+4))==(8+4)) {//already timeout despite reducing height
  var colours = new Array(3);
//defaultStatus += "FAKEX";
  var png = new Pnglet(5, 5, 3);
  colours[0] = png.color(0,0,0,0);  //Actually creates the colour palette
  colours[1] = png.color(255,255,255);
  colours[2] = png.color(1,1,1);
  for(y = 5 ; y > 0; y--) {//start at top
    png.point(colours[1],  y,5-y);
    png.point(colours[2],5-y,5-y);
  }
  dims.width  = 5;
  dims.height = 5;
//defaultStatus += "Pnglt";
  dims.nset  = new Array();
  return png;
}

 dims.width  = width;
 dims.height = height; //pixmap.maxy-y;
 return png;
}
function png_pixmap(pixmap,dims,cmap) {
  var pnglet = pnglet_data(pixmap,dims,cmap);

 //output takes lots of CPU time. On 350mHz 15478 pixels per second
 dims.nset  = new Array();
 return pnglet.output(dims.nset);
}

//http://webreference.com/programming/javascript/professional/chap21/1/6.html
//Memory Leaks and Garbage Collection (hence use of line...)

function xbm_heading(width,height) {
 return	"#define pfeiffer_width " +width+"\n"+
	"#define pfeiffer_height "+height+"\n"+
	"static char pfeiffer_bits[] = {";
}
function xbm_pixmap(pixmap,dims) {/*bitmap_output*/
 var timeout_start = new Date().getTime();
 var width  = 1+pixmap.maxx-pixmap.minx;
 var bw     = Math.ceil(width/8);
 dims.nset  = new Array();
 dims.nset[0] = 0;

 var body = "";
 var bitx =  new Array(1,2,4,8,16,32,64,128);
 var first = 1;
 var y;
 for(y = pixmap.maxy ; y >= pixmap.miny; y--) {//start at top
   if(timeout(timeout_start)) {
     timedout |= 4;
     break;
   }
   var c = new Array(bw);
   if(pixmap[y] == null) {;}
   else {
     for(var x in pixmap[y]) {
       var x_ = x-pixmap.minx;
       var cx_= x_ >> 3;
       var bx_= bitx[x_%8];
       if((c[cx_] & bx_)==0) {
         c[cx_] |= bx_;
	 dims.nset[0]++;
       }
     }//endfor
   }//endif
   //if(y==0||y==pixmap.maxy)document.writeln("xbm_pixmap y "+y+"<br>\n");
   var line = "";//reduce garbage collection load, esp Sun Netscape 4.61 
   for(var i = 0 ; i < bw ; i++) {
     if(first==0)   {line += ",\n";}
     else         first=0;
     var c_ = c[i];
     //if(y==0||y==pixmap.maxy)document.writeln("i "+i+" c_"+c_+"<br>\n");
     if(c_ == null) {line += "0x00";}
     else {
       if(c_<16)    {line += "0x0"+c_.toString(16);}
       else         {line += "0x" +c_.toString(16);}
     }
   }
   body += line;
 }
 body += "}";

 dims.width  = width;
 dims.height = pixmap.maxy-y;

 return xbm_heading(width,pixmap.maxy-y)+body;

}//end xbm_pixmap

function set_bounds(x,y,pixmap) {
  if(x<pixmap.minx) {pixmap.minx=Math.round(x);}
  if(x>pixmap.maxx) {pixmap.maxx=Math.round(x);}
  if(y<pixmap.miny) {pixmap.miny=Math.round(y);}
  if(y>pixmap.maxy) {pixmap.maxy=Math.round(y);}
}

function cmap2alpha(r,g,b) {
  //Frig all 24 bits of colour map to generate alpha channel.
  var x = (r ^ g) ^ b;
  var a = 0;
  var p = 128;
  var m = 1;
  for(var i = 0; i<8; i++) { //test lsb to msb
   if((m&x)!=0) a+=p;
   m *=2;
   p /=2.0;
  }
  a = a/255;
  //alert("cmap2alpha_("+r+","+g+","+b+") x="+x+" x1="+x1+" x2="+x2+" x3="+x3+" a="+a+" p="+p+" m="+m);
  return a;
}

/*from http://astronomy.swin.edu.au/~pbourke/fractals/lsys/
Symbols The following characters have a geometric interpretation.

Character        Meaning
   F	         Move forward by line length drawing a line
   f	         Move forward by line length without drawing a line
   +	         Turn left by turning angle
   -	         Turn right by turning angle
   |	         Reverse direction (ie: turn by 180 degrees)
   [	         Push current drawing state onto stack
   ]	         Pop current drawing state from the stack
   #	         Increment the line width by line width increment
   !	         Decrement the line width by line width increment
   @	         Draw a dot with line width radius
   {	         Open a polygon
   }	         Close a polygon and fill it with fill colour
   >	         Multiply the line length by the line length scale factor
   <	         Divide the line length by the line length scale factor
   &	         Swap the meaning of + and -
   (	         Decrement turning angle by turning angle increment
   )	         Increment turning angle by turning angle increment

WBL
& will not be implemented - used as separator
% cutoff
*/
function evalstring(params,commands,cmap,svgcomment) {
    //alert("evalstring("+params.v+" "+params.side+","+commands.length+","+cmap.length+","+svgcomment+")");
    var timeout_start = new Date().getTime();
    var time=0;

    var ncolours = cmap.length;
    if(!ncolours) ncolours =2; //black(transparent),white
    else          ncolours+=2;

    if(svgcomment) {
    var colours = new Array(ncolours);
    var alphas  = new Array(ncolours);
    colours[0] = "black";
    colours[1] = "white";
    alphas[0]  = 1;
    alphas[1]  = 1;
    for(var k=0; k<colours.length-2;k++) {
      colours[k+2]="#"+toHex(cmap[k][0])+toHex(cmap[k][1])+toHex(cmap[k][2]); 
      alphas[k+2] =cmap2alpha(cmap[k][0],cmap[k][1],cmap[k][2]);
    }
    }//end create fill colour map for SVG

    pixmap = new Array; //y x for efficiency
    pixmap.txt="";
    pixmap.minx=0;
    pixmap.maxx=0;
    pixmap.miny=0;
    pixmap.maxy=0;

    var stack = new Array;
    var sp=0;
    stack[sp] = new Object;
    stack[sp].x = 0;
    stack[sp].y = 0;
    stack[sp].a = 0;
    stack[sp].c = 1; //white stroke
    stack[sp].c2= 1; //white fill
    stack[sp].deg = params.v; 
    stack[sp].sc  = params.side; 
    stack[sp].w   = params.w;

    var pstck = new Array();
    var psp=0;
    pstck[psp] = new Object;
    pstck[psp].x = new Array();
    pstck[psp].y = new Array();
    pstck[psp].c = stack[sp].c;
    pstck[psp].c2= stack[sp].c2;
    pstck[psp].w = stack[sp].w;

    for(var l=0; l<commands.length; l++) {
	var c = commands.charAt(l);
	switch(c) { 	//do as switch rather than eval
	    case "F": {/*move();     */
	        var angle = stack[sp].a*Math.PI/180;
		var cos_ = Math.cos(angle);
		var sin_ = Math.sin(angle);
		var newx = stack[sp].x + stack[sp].sc*cos_;
		var newy = stack[sp].y + stack[sp].sc*sin_;
		set_bounds(stack[sp].x,stack[sp].y,pixmap);//assume distance>=0
		if(svgcomment) {
		  pixmap.txt +="<g id=\"F"+l+"\">\n";
		  if(stack[sp].w>=2) {//use rotated rectangle
		  var x= stack[sp].x;
		  var y= stack[sp].y;
		  var wcos2 = stack[sp].w*Math.cos(angle+Math.PI/2);
		  var wsin2 = stack[sp].w*Math.sin(angle+Math.PI/2);
		  set_bounds(x   +wcos2,   y+wsin2,pixmap);
		  set_bounds(newx+wcos2,newy+wsin2,pixmap);
		  pixmap.txt +="<g transform=\""+
		               "translate("+x+" "+y+")"+
			       " rotate("+stack[sp].a+")\">\n";
	          pixmap.txt +="<rect class=\"F\" x=\"0\" y=\"0\""+ 
			       " width=\"" +stack[sp].sc+"\""+
			       " height=\""+stack[sp].w +"\"";
		  if(stack[sp].c2>=0)
		  pixmap.txt +=" fill=\""          +colours[stack[sp].c2]+"\""+
			       " fill-opacity=\""  +alphas [stack[sp].c2]+"\"";
		  pixmap.txt +=" stroke=\""        +colours[stack[sp].c ]+"\""+
			       " stroke-opacity=\""+alphas [stack[sp].c ]+"\""+
			       " stroke-width=\"1\" />\n";
		  pixmap.txt +="</g>\n";
		  } else if(stack[sp].w>=0) {
		  pixmap.txt +="<line class=\"F\""+		  
			       " x1=\""+stack[sp].x+"\" y1=\""+stack[sp].y+"\""+
			       " x2=\""+newx+"\" y2=\""+newy+"\""+
			       " stroke=\""        +colours[stack[sp].c ]+"\""+
			       " stroke-opacity=\""+alphas [stack[sp].c ]+"\""+
			       " stroke-width=\""  +	    stack[sp].w  +"\" />\n";
		  }
		  pixmap.txt +="</g>\n";
		  }
		else {
		/* width only supported by SVG */
		for(var i=0; i<=stack[sp].sc;i++) {
		  var x_ = Math.round(stack[sp].x+i*cos_);
		  var y_ = Math.round(stack[sp].y+i*sin_);
		  if(pixmap[y_]==null) {pixmap[y_] = new Array;}
		  pixmap[y_][x_] = stack[sp].c;
		  if((time++)%100==0 && timeout(timeout_start)) {
		    set_bounds(x_,y_,pixmap);
		    timedout |= 2;
		    return pixmap;
		  }//endif timedout
		}
		}//endif SVG
		set_bounds(newx,       newy,       pixmap);//assume distance>=0
		stack[sp].x = newx;
		stack[sp].y = newy;
		break;
		}//end move
	    case "+": stack[sp].a += stack[sp].deg; break;
	    case "-": stack[sp].a -= stack[sp].deg; break;
	    case "G": /* as f ?*/
	    case "f": {//jump();     
	        var angle = stack[sp].a*Math.PI/180;
		var cos_ = Math.cos(angle);
		var sin_ = Math.sin(angle);
	        stack[sp].x += stack[sp].sc*cos_;
		stack[sp].y += stack[sp].sc*sin_;
		break;
		}//end jump
	    case "[": /*push();*/
	       var old = sp;
	       sp++;
	       if(stack[sp]==null) stack[sp] = new Object;
	       stack[sp].x = stack[old].x;
	       stack[sp].y = stack[old].y;
	       stack[sp].a = stack[old].a;
	       stack[sp].c = stack[old].c;
	       stack[sp].c2= stack[old].c2;
	       stack[sp].deg = stack[old].deg; 
	       stack[sp].sc  = stack[old].sc; 
	       stack[sp].w   = stack[old].w;
	       break;
	    case "]": /*pop();*/
	       if(sp>0) --sp; 
	       break;
	    case "{": /*push polygon */
	       psp++;
	       pstck[psp] = new Object;
	       pstck[psp].x = new Array();
	       pstck[psp].y = new Array();
	       pstck[psp].c = stack[sp].c;
	       pstck[psp].c2= stack[sp].c2;
	       pstck[psp].w = stack[sp].w;
//defaultStatus += "p{"+psp;
	       break;
	    case ".": 
	       pstck[psp].x[pstck[psp].x.length] = Math.round(stack[sp].x);
	       pstck[psp].y[pstck[psp].y.length] = Math.round(stack[sp].y);
	       set_bounds(stack[sp].x,stack[sp].y,pixmap);//assume will be used
	       break;
	    case "}": {/*pop polygon*/
	       if(svgcomment) {
	         pixmap.txt +="<g id=\"P"+l+"\">\n";
		 pixmap.txt +="<polygon class=\"P\" points=\"";
		 for(var i = 0; i<pstck[psp].x.length;i++) {
		   pixmap.txt +=pstck[psp].x[i]+" "+pstck[psp].y[i]+",";}
		 pixmap.txt +="\"\n";
		 pixmap.txt +="style=\"";
		 if(stack[sp].c2>=0) {
		 pixmap.txt +="fill:"          +colours[stack[sp].c2]+"; ";
		 pixmap.txt +="fill-opacity:"  +alphas [stack[sp].c2]+"; ";}
		 pixmap.txt +="stroke-width:"          +stack[sp].w  +"; ";
		 pixmap.txt +="stroke:"        +colours[stack[sp].c] +"; ";
		 pixmap.txt +="stroke-opacity:"+alphas [stack[sp].c];
		 pixmap.txt +="\">\n";
		 pixmap.txt +="</polygon></g>\n";
	       }
	       else {
	       if(!pixmap.polygon) pixmap.polygon = new Array();
	       var n = pixmap.polygon.length;
	       pixmap.polygon[n] = new Object;
	       pixmap.polygon[n].line = pstck[psp].c;
	       pixmap.polygon[n].fill = stack[sp].c2,
	       pixmap.polygon[n].x    = pstck[psp].x;
	       pixmap.polygon[n].y    = pstck[psp].y;
	       }//endif SVG
//defaultStatus += " "+psp+"}"+pixmap.polygon[n].x.length+","+n;
//defaultStatus += ","+pixmap.polygon[n].line+","+pixmap.polygon[n].fill;
	       if(psp>0) --psp;
	       break;}
	    case "'": 
	       if(stack[sp].c<ncolours-1) stack[sp].c++; 
	       stack[sp].c2 = (stack[sp].c2+1)%ncolours; //fill may != c
	       break;
	    case "`": 
	       if(stack[sp].c>0) stack[sp].c--;
	       else              stack[sp].c = ncolours-1;
	       stack[sp].c2 = (stack[sp].c2-1)%ncolours;
	       break;
	    case "%": {/* cutoff */
	       var l_ = l+1;
	       var c;
	       while((l_<commands.length) &&(c = commands.charAt(l_))!="]"){
	         if(c=="}") {if(psp>0) --psp;} /*pop polygon but dont draw it*/
	         l=l_; l_++;
	       }
	       break;}
	    case "|": /* Reverse direction (ie: turn by 180 degrees) */
	       stack[sp].a += 180;
	       break;
	    case "#": /* Increment the line width by line width increment */
	       stack[sp].w *= params.wf; //use factor
	       break;
	    case "!": /* Decrement the line width by line width increment */
	       stack[sp].w /= params.wf; //use factor
	       break;
	    case ">": /* Multiply the line length by the line length scale factor */
	       stack[sp].sc *= params.sf; 
	       break;
	    case "<": /* Divide the line length by the line length scale factor */
	       stack[sp].sc /= params.sf; 
	       break;
	    case "(": /* Decrement turning angle by turning angle increment */
	       stack[sp].deg -= params.vf; 
	       break;
	    case ")": /* Decrement turning angle by turning angle increment */
	       stack[sp].deg += params.vf; 
	       break;
	    case "@": /* Draw a dot with line width radius */
		if(svgcomment) {
		  var x = stack[sp].x;
		  var y = stack[sp].y;
		  var w = stack[sp].w;
		  pixmap.txt +="<g id=\"C"+l+"\">\n";
		  pixmap.txt +="<circle"+
			       " cx=\""+x+"\""+
		               " cy=\""+y+"\""+
			       " r=\"" +w+"\"";
		  if(stack[sp].c2>=0)
		  pixmap.txt +=" fill=\""          +colours[stack[sp].c2]+"\""+
			       " fill-opacity=\""  +alphas [stack[sp].c2]+"\"";
		  pixmap.txt +=" stroke=\""        +colours[stack[sp].c ]+"\""+
			       " stroke-opacity=\""+alphas [stack[sp].c ]+"\""+
			       " stroke-width=\"1\"/>\n";
		  pixmap.txt +="</g>\n";
		  set_bounds(x-w,y-w,pixmap);
		  set_bounds(x+w,y-w,pixmap);
		  set_bounds(x-w,y+w,pixmap);
		  set_bounds(x+w,y+w,pixmap);
		}
	        else { /* width only supported by SVG. So assume w=1 */
		  var x_ = Math.round(stack[sp].x);
		  var y_ = Math.round(stack[sp].y);
		  if(pixmap[y_]==null) {pixmap[y_] = new Array;}
	          pixmap[y_][x_] = stack[sp].c;
		  set_bounds(x_,y_,pixmap);
		  } //endif SVG
	       break;
	    default: /* do nothing*/
            //missing three-d ~(predefined graphics) &
        }//end switch c
	if((time++)%100==0 && timeout(timeout_start)) {
	  timedout |= 2;
	  return headandtail(pixmap,timeout_start,svgcomment);
	}
    }//endfor
return headandtail(pixmap,timeout_start,svgcomment);
}

function headandtail(pixmap,timeout_start,svgcomment) {
if(svgcomment) {
 var width  = 1+pixmap.maxx-pixmap.minx;
 var height = 1+pixmap.maxy-pixmap.miny;
/*
pixmap.txt +="<g id=\"F"+0+"\">\n";
pixmap.txt +="<line class=\"F\" ";
pixmap.txt +="x1=\""+pixmap.minx+"\" y1=\""+pixmap.miny+"\" ";
pixmap.txt +="x2=\""+pixmap.maxx+"\" y2=\""+pixmap.maxy+"\" ";
pixmap.txt +="style=\"stroke:red\">\n";
pixmap.txt +="</line>\n";
pixmap.txt +="</g>\n";
*/
var txt="";
txt +="<?xml version=\"1.0\"?>\n";
txt +="<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.0//EN\"\n";
txt +="\"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd\">\n";
txt +="<svg xmlns=\"http://www.w3.org/2000/svg\" ";
//txt +=" width=\""+width+"px\" ";
//txt +="height=\""+height+"px\" ";
//txt +="viewBox=\"0 0 "+width+" "+height+"\";
txt +=">\n";
txt +="<title>Pfeiffer ";
txt +="http://www.cs.ucl.ac.uk/staff/W.Langdon/pfeiffer.html";
txt +="</title>\n";
txt +="<desc>Evolved Lindenmayer System</desc>\n";
//Avoid "--" confusing XML parser
txt +="<!--"+svgcomment.replace(/--/g,"__")+" -->\n";
if(timedout!=0) txt += "<!-- timedout="+timedout+" -->\n";
txt +="<!--created "+Date()+"-->\n";
var now = new Date().getTime();
txt +="<!--SVG generation took "+(now-timeout_start)+"ms -->\n";
txt +="<!--Logo bounding box "+pixmap.minx+" "+pixmap.miny+" "+pixmap.maxx+" "+pixmap.maxy+" -->\n";
txt +="<g id=\"top\" ";
txt +="transform=\"translate("+(-pixmap.minx)+","+pixmap.maxy+") ";
txt +="scale(1,-1)\">\n";
pixmap.txt = txt+pixmap.txt+"</g>\n</svg>\n";
}//endif svgcomment
    return pixmap;
}


//angle and step side used elsewhere
function D01(it,start,rules) {
    var timeout_start = new Date().getTime();
    var time=0;
    expansion=start;
    for(var i=0; i<it; i++) {
	var newstr = "";
	var size = 0;
	for(var l=0; l<expansion.length; l++) {
	    var c = expansion.charAt(l);
	    if(rules[c]!=null) c = rules[c];
	    newstr += c;
	    if((time++)%100==0 && timeout(timeout_start)) {
	      timedout |= 1;
	      return expansion;
	    }
	}
	expansion = newstr;
    }
    return expansion;
}//end D01
// output text mode colour portable bitmap $Revision: 1.3 $
// W. B. Langdon
// http://www.cs.essex.ac.uk/staff/W.Langdon
// Written on 16 April 2006 (based on lsys.js r1.25)

// --------------------
// Tested with Mozilla Rhino r1.6R2

//ppm see http://en2.wikipedia.org/wiki/Portable_Pixmap_File_Format
//        http://www-info2.informatik.uni-wuerzburg.de/mitarbeiter/wolfram/lehre/bildformate.html

function pbm_heading(width,height,comment) {
 return	"P3\n"+
	"#Generated by pfeiffer.js "+new Date()+" "+comment+"\n"+
	width+"\n"+
	height+"\n"+
	"255";
}
function print_pbm(pixmap,width,height,cmap,comment){
  print(pbm_heading(width,height,comment));

  var colours = new Array((cmap)? cmap.length+2 : 2);
  colours[0] = "0 0 0";
  colours[1] = "255 255 255";
  for(var k=0; k<colours.length-2;k++) {
    colours[k+2] = ""+cmap[k][0]+" "+cmap[k][1]+" "+cmap[k][2];}

  for(var y = height-1 ; y >= 0; y--) {//start at top
  for(var x = 0 ; x < width; x++)     {//left to right
    if(pixmap[y] && pixmap[y][x]) print(colours[pixmap[y][x]]);
    else                          print(colours[0]);
  }}
}//end print_pbm
//Javascript file to interpret one Lindenmayer grammar $Revision: 1.4 $
//and output it as a portable bit map
// W. B. Langdon
// http://www.cs.essex.ac.uk/staff/W.Langdon
// Written on 16 April 2006 Based on pfeiffer_local.html r1.113


<!-- "http://javascriptsource.com">The JavaScript Source -->
<!-- ONE STEP TO INSTALL SHIMMERING LINKS: -->
<!-- This script and many more are available free online at -->
<!-- The JavaScript Source!! http://javascript.internet.com -->
<!-- Original:  Phil Reeve  (curry_master@myrealbox.com) -->
<!-- Web Site:  http://www.philreeve.com -->
///////////////////Hex Conversions/////////////////////
// convert a single digit (0 - 16) into hex
function enHex(aDigit)
{
    return("0123456789ABCDEF".substring(aDigit, aDigit+1))
}
function toHex(n) //8bit version!
{
   return(/*enHex((0xf00000 & n) >> 20) +
            enHex((0x0f0000 & n) >> 16) +
            enHex((0x00f000 & n) >> 12) +
            enHex((0x000f00 & n) >>  8) +*/
            enHex((0x0000f0 & n) >>  4) +
            enHex((0x00000f & n) >>  0));
}

//Non dynamic parameters
eve = new Object();
eve.chrome = "v=60&str=F++F++F & it=2 & sc =5 & rules=('F','F-F++F-F')&"+ //Koch snefnug
             "w= 1&wf=1.316074&" + //3**(0.25)
	     "sf=1.732051&"      + //3**(0.5)
	     "vf=5";               //deginc

eve.v=60; eve.str=""; eve.it=0; eve.sc=5; //last ditch defaults
eve.w=1;  eve.wf =1;  eve.sf=1; eve.vf=0;
parse_chrome2(eve.chrome,eve);

function cleanchrome(text1) { //prepare to send as ?URL
//var whitespace = " \t\r\n";
  var text2='';

  for (var i=0;i<text1.length;i++) {
    var c = text1.charCodeAt(i);
    if(c > 32 && c < 128) { //discard control codes
        text2 += text1.charAt(i);
    }
  }
  return text2;
}

  function colour_map(string,F) {
//  defaultStatus = "#";
    F.colours = new Array(); 
    var c = cleanchrome(string);
    var h = c.split("#");
    for(var i=0; i<h.length; i++) {
      var s = h[i].replace(/[^0-9a-fA-F]/g,"");
      for(;s.length>0; s=s.substr(6)) {
        var n = parseInt(s.substr(0,6),16);
	if(!isNaN(n)) {
          var map = F.colours[F.colours.length] = new Array();
	  for(var k=2; k>=0; k--) { map[k] = n & 0xff; n = n >> 8;}
/*	  defaultStatus += "'"+s.substr(0,6)+"'"+"-"+
.                           map[0].toString(16)+" "+
.                           map[1].toString(16)+" "+
.                           map[2].toString(16)+" ";
*/
	}
      }
    }
//  defaultStatus += ""+F.colours.length;
  }

  function expand_chrome(newchrome,F) {
        colour_map(newchrome,F);

        parse_chrome(newchrome,F);

	F.cmds=D01(F.it,F.str,F.rules);
  }//end expand_chrome

  function embryo_log(start,err,newchrome,F,a,b,c,png_) {
    F.time = (new Date().getTime())-start;
    return cleanchrome(newchrome)+":_;"+
//	          "E="+err+";Fit="+F.fit+";"+
//		  ((a>=0)? "an="+Flake[a].n+";af="+fit[a].value+";" : "")+
//		  ((b>=0)? "bn="+Flake[b].n+";bf="+fit[b].value+";" : "")+
	          "v="+F.v+";str="+F.str+";"+
		  "it="+F.it+";sc="+F.sc+";"+
		  "w="+F.w+";wf="+F.wf+";sf="+F.sf+";vf="+F.vf+";"+
		  "rules=("+F.rules_+");"+
		  (F.str? "len="+F.str.length+";" : "")+
		  (F.cmds? "cmds="+F.cmds.length+";" : "")+
		  "elapsed="+F.time+
		  ";w="+F.width+";h="+F.height+";"+
//		  (F.nset? "nset="+F.nset.join():"")+
//		  ";to="+timedout+";"+
//		  c+";"+
//		  ((png_==null)? "": 
//		  ((png_==SVG)? "svg": png_? "png":"xbm")+"["+F.n+"]");
		  F.cmds;
  }//end embryo_log

  function embryo(u,newchrome,a,b,c) {
    var start = new Date().getTime();
    F = new Object();

    expand_chrome(newchrome,F);

    var comment =   newchrome+"\n"+
		  ((a>=0)? "an="+Flake[a].n+";af="+fit[a].value+";" : "")+
		  ((b>=0)? "bn="+Flake[b].n+";bf="+fit[b].value+";" : "")+
	          "v="+F.v+";str="+F.str+";"+
		  "it="+F.it+";sc="+F.sc+";"+
		  "w="+F.w+";wf="+F.wf+";sf="+F.sf+";vf="+F.vf+";"+
		  "rules=("+F.rules_+");"+
		  (F.str? "len="+F.str.length+";" : "")+
		  c+";"
    F.pixmap = evalstring(F,F.cmds,F.colours,"");

    F.width  = 1+F.pixmap.maxx-F.pixmap.minx;
    F.height = 1+F.pixmap.maxy-F.pixmap.miny;
    F.comment=embryo_log(start,"ok",newchrome,F,a,b,c,"");
    return F;
  }//end embryo

  function parse_rules(string,f) {
    var value = string.replace(/[\(\)\s]/g,"");
    var v_ = value.split(",");
    //text[i] += " R"+name[0]+"value="+value+" nparams="+v_.length/2;
    for(var j=0; j<v_.length;) {
      //for colour we need ' So remove ' only from even v_[j]
      //15 Mar 2004 D01 only matches 1 char strings. Note longer
      //strings have no effect. 
      f.rules[v_[j].replace(/[\']/g,"").substr(0,1)] = 
                                            (j+1<v_.length)? v_[j+1] : "";
      f.rulesset = 1;
      j += 2;
    }//endforeach part of rules
    //text[i] += " rulesset="+rulesset;
  }//end parse_rules

  function format_rules_(f) {
    f.rules_ = "";
    for(var keys in f.rules) {
      if(f.rules_!="") {f.rules_ += ",";}
      f.rules_ += keys+"=>";
      if(f.rules[keys]) f.rules_ += f.rules[keys];
    } 
  }

  function parse_chrome2(string,f) {//parse but do not supply defaults
    f.rules = new Object();
    f.rulesset=0;
    var c= cleanchrome(string);
    //as lsys.perl r1.11
    var pairs = c.split("&");
    for(pair in pairs) {
      var name = pairs[pair].split("=");
      var name1;
      var n;
      if(name.length>1) {
	name1 = name[1];
	n     = parseFloat(name[1]);
	if(isNaN(n)) n=0; //as lsys.perl
      }
      else {
	name1 =""; //as lsys.perl
	n=0;       //as lsys.perl
      }
      if(name[0]=="v"  )      {f.v  =n;}
      if(name[0]=="vf"  )     {f.vf  =n;}
      if(name[0]=="str")      {f.str=name1}
      if(name[0]=="it" )      {f.it =n;}
      if(name[0]=="sc" )      {f.sc =n;}
      if(name[0]=="sf" )      {f.sf =n;}
      if(name[0]=="w"  )      {f.w  =n;}
      if(name[0]=="wf" )      {f.wf =n;}
      if(name[0]=="rules")    {parse_rules(name1,f);}
    }//endforeach pairs
  }

  function parse_chrome(string,f) {
    f.v = eve.v; f.str = eve.str; f.it = eve.it; f.sc = eve.sc;
    f.w = eve.w; f.wf  = eve.wf;  f.sf = eve.sf; f.vf = eve.vf;
    parse_chrome2(string,f);
    if(f.rulesset==0) {f.rules = eve.rules;}
    format_rules_(f);

    f.side = f.sc*0.2;
  }

var limit=0; //zero disables it
var timedout;
  function timeout(timeout_start) {
    return false;
}
function print_help() {
  print("One shot version of http://www.cs.ucl.ac.uk/staff/W.Langdon/pfeiffer.html")
  print("Converts one Lindenmayer grammar into a portable bit map graphics file");
  print("First command line argument file holding grammar or grammar");
  print("P3 picture written to standard out");
  print("pfeiffer.js $Revision: 1.4 $");
}

if(arguments.length==0) print_help();
else {
  //Try and read grammar from file. If that fails use command line
  var t = readFile(arguments[0]);
  var chrome = (t.length>0)? t : arguments[0];

  //Grow grammar into graphics commands and then into picture
  var Flake = embryo(false,chrome,-1,-1,0);

  //Convert picture and then output it as pbm
  //Since Pnglets implements polygon we use it to generate sparse pixmap
  var pnglet = pnglet_data(Flake.pixmap,Flake,Flake.colours);

  var comment = "pfeiffer.js $Revision: 1.4 $ "+arguments[0]+" "+
                t.replace(/\n/g,"")+";"+F.comment;
  print_pbm(pnglet.data,Flake.width,Flake.height,Flake.colours,comment);
}
